
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: _G, coroutine, IsAddOnLoaded, Vd1, VUHDO_CONFIG, VUHDO_getDynamicModelArray, VUHDO_getGroupMembers, VUHDO_PANEL_SETUP, VUHDO_RAID, VUHDO_redrawAllPanels, VUHDO_vuhdoInitLocalOverrides

local format = format
local InCombatLockdown = InCombatLockdown
local max = max
local min = min
local next = next
local unpack = unpack

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Integration VuhDo")

local PANEL = 1

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

function M:OnEnable()
	if not IsAddOnLoaded("VuhDo") then
		self:RegisterEvent("ADDON_LOADED")
		return
	end

	if not self:IsHooked("VUHDO_determineRole") then
		self:RawHook("VUHDO_determineRole")
	end

	if not self:IsHooked("VUHDO_getBarIcon") then
		self:RawHook("VUHDO_getBarIcon")
	end

	if not self:IsHooked("VUHDO_getBarIconFrame") then
		self:RawHook("VUHDO_getBarIconFrame")
	end

	if not self:IsHooked("VUHDO_getOrCreateActionPanel") then
		self:SecureHook("VUHDO_getOrCreateActionPanel")
	end

	if not self:IsHooked("VUHDO_getOrCreateHealButton") then
		self:SecureHook("VUHDO_getOrCreateHealButton")
	end

	if VUHDO_CONFIG then
		VUHDO_vuhdoInitLocalOverrides()
	end

	self:RegisterMessage("SIXUI_CONFIG_UPDATE")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED", "Update")
	self:RegisterMessage("SIXUI_PROFILE_UPDATE", "Update")
	self:StyleButtons()
end

function M:OnDisable()
	self:UnstyleButtons()
end

function M:Update()
	self:AdjustButtons()
	self:AdjustPrimaryPanel()
end

function M:IterateButtons()
	return coroutine.wrap(function()
		for panel = 1, 10 do
			for button = 1, 51 do
				local button = _G[format("Vd%dH%d", panel, button)]

				if button then
					coroutine.yield(button)
				end
			end
		end
	end)
end

-- Create borders for all healing buttons.
function M:StyleButtons()
	for button in self:IterateButtons() do
		self:StyleButton(button)
	end
end

-- Apply SixUI's style to a healing button.
function M:StyleButton(button)
	if (not button) then return end

	if not self:IsHooked(button, "Show") then
		self:SecureHook(button, "Show", "Button_OnShow")
	end

	if not self:IsHooked(button, "Hide") then
		self:SecureHook(button, "Hide", "Button_OnHide")
	end

	self[button:IsShown() and "Button_OnShow" or "Button_OnHide"](self, button)
end

-- Remove borders from all healing buttons.
function M:UnstyleButtons()
	for button in self:IterateButtons() do
		self:UnstyleButton(button)
	end
end

function M:UnstyleButton(button)
	self:Unhook(button, "OnHide")
	self:Unhook(button, "OnShow")
	self:Button_OnHide(button)
end

-- Adjust the profile of all healing buttons based on the group size.
function M:AdjustButtons()
	if InCombatLockdown() then return end
	local raidFrameConfig = self:GetRaidFrameConfig()
	if self:HasBarSize(raidFrameConfig) then return end

	for panel = 1, 10 do
		VUHDO_PANEL_SETUP[panel]["SCALING"]["barWidth"] = raidFrameConfig.width
		VUHDO_PANEL_SETUP[panel]["SCALING"]["barHeight"] = raidFrameConfig.height
	end

	VUHDO_redrawAllPanels()
end

-- Adjust the position of the primary panel.
function M:AdjustPrimaryPanel()
	if InCombatLockdown() then return end
	local raidFrameConfig = self:GetRaidFrameConfig()
	if self:HasPanelPosition(raidFrameConfig) then return end

	A:RegisterFrame("Vd1"):ClearAllPoints():SetPoint(raidFrameConfig.point, raidFrameConfig.relativeTo, raidFrameConfig.relativePoint, raidFrameConfig.offsetX, raidFrameConfig.offsetY)
end

-- Determine how many healing buttons are being displayed.
function M:GetNumButtonsPanel(panel)
	local model = VUHDO_getDynamicModelArray(panel)
	local count = 0

	for index, id in next, model do
		count = count + #VUHDO_getGroupMembers(id, panel, index)
	end

	return count
end

-- Determine the most relevant raid frame config.
function M:GetRaidFrameConfig()
	return A:GetRaidFrameConfig(self:GetNumButtonsPanel())
end

-- Determine whether a raid frame config's bar size is active.
function M:HasBarSize(raidFrameConfig)
	return VUHDO_PANEL_SETUP[PANEL]["SCALING"]["barWidth"] == raidFrameConfig.width and VUHDO_PANEL_SETUP[PANEL]["SCALING"]["barHeight"] == raidFrameConfig.height
end

-- Determine whether a raid frame config's position is active.
function M:HasPanelPosition(raidFrameConfig)
	local point, relativeTo, relativePoint, offsetX, offsetY = Vd1:GetPoint()

	return raidFrameConfig.point == point
		and raidFrameConfig.relativeTo == relativeTo:GetName()
		and raidFrameConfig.relativePoint == relativePoint
		and raidFrameConfig.offsetX == offsetX
		and raidFrameConfig.offsetY == offsetY
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:ADDON_LOADED(event, name)
	if name == "VuhDo" then
		self:UnregisterEvent("ADDON_LOADED")
		self:OnEnable()
	end
end

function M:SIXUI_CONFIG_UPDATE(event, module, config, key, ...)
	if module == "Core RaidFrameConfig" then
		if key == "width" or key == "height" then
			self:AdjustButtons()
		else
			self:AdjustPrimaryPanel()
		end
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

-- Use a hook to set SixUI's border color to VuhDo's bar border color.
function M:Border_Hide(border)
	if border.GetBorder then
		border:GetBorder():SetColor()
	end
end

-- Use a hook to set SixUI's border color to VuhDo's bar border color.
function M:Border_Show(border)
	if border.GetBorder then
		border:GetBorder():SetColor(border:GetBackdropBorderColor())
	end
end

function M:Button_OnHide(button)
	local border = _G[format("%sBgBarIcBarHlBarPlTg", button:GetName())]

	if border then
		self:Unhook(border, "Hide")
		self:Unhook(border, "Show")
	end

	B:RemoveBorder(button)
end

function M:Button_OnShow(button)
	local border = _G[format("%sBgBarIcBarHlBarPlTg", button:GetName())]

	B:CreateBorder(button, C.BORDER_PADDING):SetBorderStrata("HIGH")

	if border then
		border.GetBorder = button.GetBorder

		if not self:IsHooked(border, "Show") then
			self:SecureHook(border, "Show", "Border_Show")
		end

		if not self:IsHooked(border, "Hide") then
			self:SecureHook(border, "Hide", "Border_Hide")
		end

		self[border:IsShown() and "Border_Show" or "Border_Hide"](self, border)
	end
end

-- Use a hook to scale debuff icons to the size of HoT icons.
function M:SetScale(tFrame, ...)
	local sHeight = max(0, min(150, VUHDO_PANEL_SETUP[1]["SCALING"]["barHeight"]))
	local sHotIconSize = max(0, min(100, VUHDO_PANEL_SETUP[1]["HOTS"]["size"]))
	local tScale = (sHeight * sHotIconSize * 0.01) / sHeight

	self.hooks[tFrame].SetScale(tFrame, tScale)
end

-- Use a hook to zoom icons by 30%.
function M:SetTexCoord(anIcon, ...)
	if ... == 0 then
		self.hooks[anIcon].SetTexCoord(anIcon, unpack(C.ZOOM_TEX_COORD))
	else
		self.hooks[anIcon].SetTexCoord(anIcon, ...)
	end
end

-- Use a hook to override how VuhDo sorts classes.
function M:VUHDO_determineRole(aUnit)
	local tInfo = VUHDO_RAID[aUnit]
	local tClassId = tInfo["classId"]
	local tRole = self.hooks["VUHDO_determineRole"](aUnit)

	-- VUHDO_ID_HUNTERS
	if self.config.overrideHunterRole and tClassId == 22 then
		return 62 -- VUHDO_ID_RANGED_DAMAGE
	end

	 -- VUHDO_ID_PALADINS
	if self.config.overrideHolyPaladinRole and tClassId == 23 and tRole == 63 then
		return 61 -- VUHDO_ID_MELEE_DAMAGE
	end

	return tRole
end

-- Use a hook to find bar icons.
function M:VUHDO_getBarIcon(...)
	local anIcon = self.hooks["VUHDO_getBarIcon"](...)

	if anIcon and not self:IsHooked(anIcon, "SetTexCoord") then
		self:RawHook(anIcon, "SetTexCoord", true)
		self:SetTexCoord(anIcon, anIcon:GetTexCoord())
	end

	return anIcon
end

-- Use a hook to find bar icon frames.
function M:VUHDO_getBarIconFrame(aButton, anIconNumber)
	local tFrame = self.hooks["VUHDO_getBarIconFrame"](aButton, anIconNumber)

	if tFrame and anIconNumber >= 40 and not self:IsHooked(tFrame, "SetScale") then
		self:RawHook(tFrame, "SetScale", true)
		self:SetScale(tFrame, tFrame:GetScale())
	end

	return tFrame
end

-- Use a hook to find action panels.
function M:VUHDO_getOrCreateActionPanel(aPanelNum)
	A.FrameQueue:Process(format("Vd%s", aPanelNum))
	self:AdjustPrimaryPanel()
end

-- Use a hook to find heal buttons.
function M:VUHDO_getOrCreateHealButton(button, panel)
	self:StyleButton(_G[format("Vd%dH%d", panel, button)])
	self:AdjustButtons()
end
