
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: ipairs, IsAddOnLoaded, pairs, table, WeakAuras, WeakAurasSaved

local coroutine = coroutine
local next = next
local unpack = unpack

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Integration WeakAuras")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.RegionType = {
	AuraBar = "aurabar",
	SubBorder = "subborder",
	Icon = "icon",
}

M.RegionTypeWithStyle = {
	[M.RegionType.AuraBar] = true,
	[M.RegionType.Icon] = true,
}

M.originalTexture = {}

M.regionRegistry = {}

function M.GetColorForBorder(border)
	if border and border:IsShown() then
		local r, g, b, a = border:GetBackdropBorderColor()

		if (r ~= 0 or g ~= 0 or b ~= 0) and (r ~= 1 or g ~= 1 or b ~= 1) and a ~= 0 then
			return r, g, b
		end
	end
end

function M.GetKeyInfo(t)
	local keys = {}

	for key in pairs(t) do
		table.insert(keys, key)
	end

	return keys, #keys, 0
end

function M:OnEnable()
	if not IsAddOnLoaded("WeakAuras") then
		self:RegisterEvent("ADDON_LOADED")
		return
	end

	if not self:IsHooked(WeakAuras.regionPrototype, "create") then
		self:SecureHook(WeakAuras.regionPrototype, "create", "RegionPrototype_create")
	end

	if not self:IsHooked(WeakAuras.regionPrototype, "modifyFinish") then
		self:SecureHook(WeakAuras.regionPrototype, "modifyFinish", "RegionPrototype_modifyFinish")
	end

	for region in self:IterateRegions() do
		self:RegisterRegion(region)
	end

	self:RegisterMessage("SIXUI_CLUTTER_STATE_CHANGED")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED")
end

function M:OnDisable()
	for region in self:IterateRegions() do
		self:UnregisterRegion(region)
	end
end

-- Iterate over known and discoverable WeakAuras regions.
function M:IterateRegions()
	local keys = self.GetKeyInfo(WeakAurasSaved and WeakAurasSaved.displays or {})

	for _, key in ipairs(keys) do
		local region = WeakAuras.GetRegion(key)

		if region then
			self.regionRegistry[region] = true
		end
	end

	return coroutine.wrap(function()
		for region in next, self.regionRegistry do
			if region and region.Show and region.Hide then
				coroutine.yield(region)
			end
		end
	end)
end

-- Iterate over sub-regions of a given type.
function M:IterateSubRegions(subRegions, subRegionType)
    local subRegionIndex = 0

	return coroutine.wrap(function()
        for _, subRegion in next, subRegions do
            if not subRegionType or subRegionType == subRegion.type then
                subRegionIndex = subRegionIndex + 1
                coroutine.yield(subRegion, subRegionIndex)
            end
        end
    end)
end

-- Register the given region for updates and style it.
function M:RegisterRegion(region)
	if not self:IsHooked(region, "OnShow") then
		self:HookScript(region, "OnShow", "Region_OnShow")
	end

	if not self:IsHooked(region, "OnHide") then
		self:HookScript(region, "OnHide", "Region_OnHide")
	end

	if region:IsShown() then
		self:Region_OnShow(region)
	else
		self:Region_OnHide(region)
	end
end

-- Unregister the given region from updates and reset it.
function M:UnregisterRegion(region)
	self:Unhook(region, "OnShow")
	self:Unhook(region, "OnHide")
	self:ResetRegion(region)
end

function M:StyleRegion(region)
	if (not self.RegionTypeWithStyle[region.regionType]) then
		return
	end

	if (region.noSixBorder) then
		return
	end

	B:CreateBorder(region, C.BORDER_PADDING)

	if region.regionType == self.RegionType.AuraBar then
		self:StyleBar(region.bar)
	end

	self:StyleSubRegions(region.subRegions)
end

function M:ResetRegion(region)
	B:RemoveBorder(region)

	if region.regionType == self.RegionType.AuraBar then
		self:ResetBar(region.bar)
	end

	self:ResetSubRegions(region.subRegions)
end

function M:StyleBar(bar)
	if (not bar) then return end
	self:StyleTexture(bar.bg, nil)
	self:StyleTexture(bar.fg, [[Interface\AddOns\SixUI\Resources\Textures\Light.tga]])
end

function M:ResetBar(bar)
	if (not bar) then return end
	self:ResetTexture(bar.bg)
	self:ResetTexture(bar.fg)
end

function M:StyleTexture(object, texture)
	if (not object) then return end

	if not self.originalTexture[object] then
		self.originalTexture[object] = {object:GetTexture()}
	end

	object:SetTexture(texture)
end

function M:ResetTexture(object)
	if (not object) then return end
	local texture = self.originalTexture[object]
	if (not texture) then return end
	object:SetTexture(unpack(texture))
	self.originalTexture[object] = nil
end

function M:StyleSubRegions(subRegions)
	if (not subRegions) then return end

	for border, borderIndex in self:IterateSubRegions(subRegions, self.RegionType.SubBorder) do
		border:SetAlpha(0)

		if borderIndex == 1 then
			if not self:IsHooked(border, "SetBorderColor") then
				self:SecureHook(border, "SetBorderColor", "Border_SetBorderColor")
			end

			if not self:IsHooked(border, "SetVisible") then
				self:SecureHook(border, "SetVisible", "Border_SetBorderColor")
			end

			self:Border_SetBorderColor(border)
		end
	end
end

function M:ResetSubRegions(subRegions)
	if (not subRegions) then return end

	for border in self:IterateSubRegions(subRegions, self.RegionType.SubBorder) do
		border:SetAlpha(1)
		self:Unhook(border, "SetBorderColor")
		self:Unhook(border, "SetVisible")
	end
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:ADDON_LOADED(event, name)
	if name == "WeakAuras" then
		self:UnregisterEvent("ADDON_LOADED")
		self:OnEnable()
	end
end

function M:SIXUI_CLUTTER_STATE_CHANGED(...)
	WeakAuras.ScanEvents(...)
end

function M:SIXUI_LAYOUT_CHANGED(...)
	WeakAuras.ScanEvents(...)
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M:Border_SetBorderColor(region)
	local border = B:GetBorder(region:GetParent())
	if (not border) then return end
	border:SetColor(self.GetColorForBorder(region))
end

function M:Region_OnHide(region)
	self:ResetRegion(region)
end

function M:Region_OnShow(region)
	self:StyleRegion(region)
end

function M:RegionPrototype_create(region)
	self.regionRegistry[region] = true
	self:RegisterRegion(region)
end

function M:RegionPrototype_modifyFinish(_, region)
	self.regionRegistry[region] = true
	self:RegisterRegion(region)
end
